@@ -5,7 +5,6 @@ import sys
 
 import numpy as np
 import pandas as pd
-import tensorflow as tf
 
 
 class Gen_Basis_Order:
@@ -187,7 +186,7 @@ class MultiQubitSystem:
 
 class Ideal_Measurements:
 
-    def __init__(self, qs=2):
+    def __init__(self, qs=2, basis_set='nisq'):
         self._qs = qs
         if not os.path.exists(f'./projectors/projectors_array_qs_{self._qs}.pickle'):
             print('| To accelerate the simulation, a new projector folder is created.')
@@ -196,7 +195,13 @@ class Ideal_Measurements:
             except:
                 print('|Found projectors folder! Appending some files ...')
             mqs = MultiQubitSystem(qs=self._qs)
-            proj = mqs.NISQ_Projectors()
+            if basis_set == 'general':
+                proj = mqs.General_Scheme_Projectors()
+            elif basis_set == 'nisq':
+                proj = mqs.NISQ_Projectors()
+            else:
+                raise ValueError('basis_set either be general or nisq.')
+            # proj = mqs.NISQ_Projectors()
             with open(f'./projectors/projectors_array_qs_{self._qs}.pickle', 'wb') as f:
                 pkl.dump(proj, f, -1)
             self.projectors = proj
@@ -246,7 +251,9 @@ class Ideal_Measurements:
 
 class Random_Measurements:
 
-    def __init__(self, qs=2, n_meas=1024):
+    ''' This is same as the Random_Measurements_Old but is much more faster.'''
+
+    def __init__(self, qs=2, n_meas=1024, basis_set='general'):
         self._qs = qs
         self.n_shots = n_meas
 
@@ -254,7 +261,12 @@ class Random_Measurements:
             print(
                 '| To accelerate the simulation, General Scheme Projector file is created in a new projectors folder.')
             mqs = MultiQubitSystem(qs=self._qs)
-            proj = mqs.General_Scheme_Projectors()
+            if basis_set == 'general':
+                proj = mqs.General_Scheme_Projectors()
+            elif basis_set == 'nisq':
+                proj = mqs.NISQ_Projectors()
+            else:
+                raise ValueError('basis_set either be general or nisq.')
             try:
                 os.mkdir('projectors')
             except:
@@ -287,50 +299,38 @@ class Random_Measurements:
         tau = np.concatenate([tau_first, tau_second])
         return tau
 
-    def get_tomo_from_density_matrix(self, projectors):
-        tomo = np.trace(np.real(np.matmul(self.rho, projectors))).astype(np.float32)  # one by one multiplication
+    def get_tomo_from_density_matrix(self, density_matrix, projectors):
+
+        tomo = np.real(np.trace(np.matmul(density_matrix, projectors), axis1=1, axis2=2))
         return tomo
 
-    def measure_specific_basis(self, proj_int, dm_one):
-        self.rho = dm_one
-        # n_proj = self.projectors.reshape(3 ** self._qs, self._qs ** 2, self._qs ** 2, self._qs ** 2)
-        projectors = self.n_proj[proj_int]  # --> [4, 4, 4]
-        tomo_probs = list(map(self.get_tomo_from_density_matrix, projectors))
-        tomo_probs = np.array(tomo_probs).flatten()
-        draw = np.random.multinomial(1, tomo_probs)
-        arg_max_rank = np.argmax(draw)
-        measurements = np.zeros(6 ** self._qs).astype(np.float32)
-        measurements[2**self._qs * proj_int + arg_max_rank] = 1.
-        self.proj_used_rank_list.append(4 * proj_int + arg_max_rank)
-        return measurements
-
     def measurement_array(self, dm=None):
-        dm = np.squeeze(dm)
+        dm = dm.reshape(1, 2**self._qs, 2**self._qs)
+        tomo_ideal = self.get_tomo_from_density_matrix(density_matrix=dm, projectors=self.projectors)
+        tomo_ideal_array = tomo_ideal.reshape(3**self._qs, 2**self._qs)
         proj_rank = np.random.randint(0, 3**self._qs, self.n_shots)
-        measurement_list = list(map(lambda x: self.measure_specific_basis(x, dm), proj_rank))
-        measurement_array = np.array(measurement_list).reshape(-1, 6 ** self._qs)
-        self.count += 1
-        if self.count % 1000 == 0:
-            print(f'dm_finished: {self.count}')
-        return measurement_array
+        occurance_rank = [np.count_nonzero(proj_rank==i) for i in range(3**self._qs)]
+        tomo_list = [np.random.multinomial(occurance_rank[i], tomo_ideal_array[i]) for i in range(3**self._qs)]
+        tomo_array = np.concatenate(tomo_list)
+        return tomo_array
 
     def tomography_data(self, density_matrix, norm=True, save_file=False, filename='pickle.pickle'):
         self.count = 1
         measurements = list(map(self.measurement_array, density_matrix))
-        measurements = np.array(measurements).reshape(-1, self.n_shots, 6 ** self._qs)
+        measurements = np.array(measurements).reshape(density_matrix.shape[0], 6 ** self._qs)
         if norm:
-            tomo_array = np.sum(measurements, axis=1) / self.n_shots
+            tomo_array = measurements / self.n_shots
         else:
-            tomo_array = np.sum(measurements, axis=1)
+            tomo_array = measurements
         tau_list = list(map(self.get_tau_cholesky, density_matrix))
         tau_array = np.array(tau_list).reshape(-1, 2 ** self._qs * 2 ** self._qs)
 
         if save_file:
             if not os.path.exists('./data/data_measurements'):
                 os.mkdir('./data/data_measurements')
-                tf.print('./data/data_measurements folder has been created and ')
-            with open(f'./data/measurements/{filename}', 'wb') as f:
-                tf.print(f'tomography data has been saved into ./data/data_measurements/{filename}')
+                print('./data/data_measurements folder has been created and ')
+            with open(f'./data/data_measurements/{filename}', 'wb') as f:
+                print(f'tomography data has been saved into ./data/data_measurements/{filename}')
                 pkl.dump([tomo_array, tau_array, density_matrix, self.proj_used_rank_list], f, -1)
 
         # return measurements, self.proj_used_rank_list
@@ -339,14 +339,20 @@ class Random_Measurements:
 
 class NISQ_Shots:
 
-    def __init__(self, qs=2, shots=1024):
+    def __init__(self, qs=2, shots=1024, basis_set='nisq'):
         self._qs = qs
         self.n_shots = shots
 
         if not os.path.exists(f'./projectors/projectors_array_qs_{self._qs}.pickle'):
             print('| To accelerate the simulation, Projectors file is created in a new projectors folder.')
             mqs = MultiQubitSystem(qs=self._qs)
-            proj = mqs.NISQ_Projectors()
+            if basis_set == 'general':
+                proj = mqs.General_Scheme_Projectors()
+            elif basis_set == 'nisq':
+                proj = mqs.NISQ_Projectors()
+            else:
+                raise ValueError('basis_set either be general or nisq.')
+            # proj = mqs.NISQ_Projectors()
             try:
                 os.mkdir('projectors')
             except:
@@ -411,10 +417,108 @@ class NISQ_Shots:
         if save_file:
             if not os.path.exists('./data/data_shots'):
                 os.mkdir('./data/data_shots')
-                tf.print('./data/data_shots folder has been created and ')
+                print('./data/data_shots folder has been created and ')
             with open(f'./data/data_shots/{filename}', 'wb') as f:
-                tf.print(f'tomography data has been saved into ./data/data_shots/{filename}')
+                print(f'tomography data has been saved into ./data/data_shots/{filename}')
                 pkl.dump([tomo_array, tau_array, density_matrix, self.proj_used_rank_list], f, -1)
 
         # return measurements, self.proj_used_rank_list
         return tomo_array, tau_array
+
+class Random_Measurements_Old:
+
+    def __init__(self, qs=2, n_meas=1024, basis_set='general'):
+        self._qs = qs
+        self.n_shots = n_meas
+
+        if not os.path.exists(f'./projectors/projectors_array_qs_{self._qs}_general_scheme.pickle'):
+            print(
+                '| To accelerate the simulation, General Scheme Projector file is created in a new projectors folder.')
+            mqs = MultiQubitSystem(qs=self._qs)
+            if basis_set == 'general':
+                proj = mqs.General_Scheme_Projectors()
+            elif basis_set == 'nisq':
+                proj = mqs.NISQ_Projectors()
+            else:
+                raise ValueError('basis_set either be general or nisq.')
+            try:
+                os.mkdir('projectors')
+            except:
+                print('|Found projectors folder! Appending some files ...')
+
+            with open(f'projectors_array_qs_{self._qs}_general_scheme.pickle', 'wb') as f:
+                pkl.dump(proj, f, -1)
+            self.projectors = proj
+        else:
+            self.projectors = pd.read_pickle(f'./projectors/projectors_array_qs_{self._qs}_general_scheme.pickle')
+
+        self.n_proj = self.projectors.reshape(3 ** self._qs, 2**self._qs , 2**self._qs, 2**self._qs)
+        self.proj_used_rank_list = []
+
+    def get_tau_cholesky(self, density_matrix):
+        try:
+            chol = np.linalg.cholesky(density_matrix)
+        except:
+            dm = (1. - 0.0000001) * density_matrix + 0.0000001 / 4 * np.eye(2 ** self._qs)
+            chol = np.linalg.cholesky(dm)
+            pass
+        tau_first = np.real(chol.diagonal())  # len 16 for 4 qubits
+        ele_off_diags = chol[np.tril_indices(n=2 ** self._qs,
+                                             k=-1)]  # https://numpy.org/doc/stable/reference/generated/numpy.tril_indices.html
+        tau_second = []
+        for i in ele_off_diags:
+            re, img = np.real(i), np.imag(i)
+            tau_second.append(re)
+            tau_second.append(img)
+        tau = np.concatenate([tau_first, tau_second])
+        return tau
+
+    def get_tomo_from_density_matrix(self, projectors):
+        tomo = np.trace(np.real(np.matmul(self.rho, projectors))).astype(np.float32)  # one by one multiplication
+        return tomo
+
+    def measure_specific_basis(self, proj_int, dm_one, proj_rank_list=False):
+        self.rho = dm_one
+        # n_proj = self.projectors.reshape(3 ** self._qs, self._qs ** 2, self._qs ** 2, self._qs ** 2)
+        projectors = self.n_proj[proj_int]  # --> [4, 4, 4]
+        tomo_probs = list(map(self.get_tomo_from_density_matrix, projectors))
+        tomo_probs = np.array(tomo_probs).flatten()
+        draw = np.random.multinomial(1, tomo_probs)
+        arg_max_rank = np.argmax(draw)
+        measurements = np.zeros(6 ** self._qs).astype(np.float32)
+        measurements[2**self._qs * proj_int + arg_max_rank] = 1.
+        if proj_rank_list:
+            self.proj_used_rank_list.append(2**self._qs * proj_int + arg_max_rank)
+        return measurements
+
+    def measurement_array(self, dm=None):
+        dm = np.squeeze(dm)
+        proj_rank = np.random.randint(0, 3**self._qs, self.n_shots)
+        measurement_list = list(map(lambda x: self.measure_specific_basis(x, dm), proj_rank))
+        measurement_array = np.array(measurement_list).reshape(-1, 6 ** self._qs)
+        self.count += 1
+        if self.count % 1000 == 0:
+            print(f'dm_finished: {self.count}')
+        return measurement_array
+
+    def tomography_data(self, density_matrix, norm=True, save_file=False, filename='pickle.pickle'):
+        self.count = 1
+        measurements = list(map(self.measurement_array, density_matrix))
+        measurements = np.array(measurements).reshape(-1, self.n_shots, 6 ** self._qs)
+        if norm:
+            tomo_array = np.sum(measurements, axis=1) / self.n_shots
+        else:
+            tomo_array = np.sum(measurements, axis=1)
+        tau_list = list(map(self.get_tau_cholesky, density_matrix))
+        tau_array = np.array(tau_list).reshape(-1, 2 ** self._qs * 2 ** self._qs)
+
+        if save_file:
+            if not os.path.exists('./data/data_measurements'):
+                os.mkdir('./data/data_measurements')
+                print('./data/data_measurements folder has been created and ')
+            with open(f'./data/measurements/{filename}', 'wb') as f:
+                print(f'tomography data has been saved into ./data/data_measurements/{filename}')
+                pkl.dump([tomo_array, tau_array, density_matrix, self.proj_used_rank_list], f, -1)
+
+        # return measurements, self.proj_used_rank_list
+        return tomo_array, tau_array
\ No newline at end of file
