@@ -1,7 +1,10 @@
+"""
+author: Sanjaya Lohani
+email: slohani@mlphys_nightly.com
+Licence: Apache-2.0
+"""
 import os
-
 import tensorflow as tf
-
 from mlphys.deepqis.utils import Extract_Net as arlnet
 
 
